#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['vgmusic']

package_data = \
{'': ['*']}

install_requires = \
['beautifulsoup4>=4.9.1', 'html5lib>=1.1', 'requests>=2.24.0']

extras_require = \
{'cli': ['click>=7.1.2'],
 'doc': ['pydoc-markdown>=3.10.1'],
 'rest': ['fastapi>=0.63.0', 'uvicorn[standard]>=0.13.3']}

entry_points = \
{'console_scripts': ['vgmusic = vgmusic.cli:cli']}

setup(name='vgmusic',
      version='1.0.2',
      description='Unofficial Python API for vgmusic.com.',
      author='Ong Yong Xin',
      author_email='ongyongxin2020+github@gmail.com',
      url='https://github.com/ongyx/vgmusic.py',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      entry_points=entry_points,
      python_requires='>=3.6',
     )
